/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Template to implement metadata format handlers.
 */

#define	HANDLER	"template"

#include "internal.h"
#define	FORMAT_HANDLER
#include "template.h"
#include <datastruct/byteorder.h>

static const char *handler = HANDLER;

/* Make up RAID device name. */
/* CODEME: implement creation of senseful name for the RAID device */
static size_t _name(struct template *template, char *str, size_t len,
		    unsigned int subset)
{
	return snprintf(str, len, "template") + 1;
}

static char *name(struct lib_context *lc, struct raid_dev *rd,
		  unsigned int subset)
{
	size_t len;
	char *ret;
	struct template *template = META(rd, template);

	if ((ret = dbg_malloc((len = _name(template, NULL, 0, subset)) + 1))) {
		_name(template, ret, len, subset);
		mk_alpha(lc, ret + HANDLER_LEN, len - HANDLER_LEN);
	} else
		log_string_err(lc, handler);

	return ret;
}

/* Mapping of template types to generic types */
/*
 * CODEME: mappings of template private level types to generic ones 
 * (see metadata.h for generic ones)
 */
static struct types types[] = {
	{ TEMPLATE_T_SPAN,  t_linear },
	{ TEMPLATE_T_RAID0, t_raid0 },
        { 0, t_undef}
};

/* Neutralize disk type using generic metadata type mapping function */
static enum type template_type(struct lib_context *lc, struct raid_dev *rd)
{
	return rd_type(types, (unsigned int) (META(rd, template))->type);
}

/*
 * Read a Template RAID device
 */
/* CODEME: endianess conversion */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct template *template)
{
	CVT32(template->something);
	...
}
#endif

static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct template *template, struct dev_info *di);
static struct raid_dev *template_read(struct lib_context *lc,
					struct dev_info *di)
{
	struct template *template;
	struct raid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata and read it. */
	/* CODEME: calculate TEMPLATE_CONFIGOFFSET correctly */
	if (!(template = alloc_private_and_read(lc, handler, sizeof(*template),
					   di->path, TEMPLATE_CONFIGOFFSET)))
		goto out;

	to_cpu(template);

	/* CODEME: unique identification of template metadata */
	if (strncmp(template->magic_name, TEMPLATE_MAGIC_NAME,
		    sizeof(TEMPLATE_MAGIC_NAME) - 1))
		goto bad;
	
	/* File metadata for tests (writes ${devicename}.dat file) */
	file_metadata(lc, handler, di->path, template, sizeof(*template),
		      TEMPLATE_CONFIGOFFSET);

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_raid_dev(lc, handler)))
		goto bad;

	/*
	 * Setup generic RAID device structure
	 * members using ondisk information.
	 */
	if (setup_rd(lc, rd, template, di))
		goto out;

	log_err(lc, "%s: setting up RAID device %s", handler, di->path);
	free_raid_dev(lc, &rd);
	goto out;

   bad:
	dbg_free(template);
   out:
	return rd;
}

/*
 * Decide about ordering sequence of RAID device.
 * (Called by list_add_dev_sorted().
 */
static int dev_sort(struct list_head *pos, struct list_head *new)
{
	return (META(RD(new), template))->disk_number <
	       (META(RD(pos), template))->disk_number;
}

/* Add a Template RAID device to a set */
static struct raid_set *template_group(struct lib_context *lc,
					 struct raid_dev *rd)
{
	/* CODEME: add grouping logic */
	log_err(lc, "%s: implement grouping logic for RAID set", handler);
	return NULL;
}

/* CODEME: Write private RAID metadata to device */
static int template_write(struct lib_context *lc,
			  struct raid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct template *template = META(rd, template);
#endif

	/* CODEME: in case there's more complex metadata areas */
	to_disk(template);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(template);
	return ret;
}

/*
 * Check integrity of a RAID set.
 */
static int template_check(struct lib_context *lc, struct raid_set *rs)
{
	/* CODEME: implement */
	log_err(lc, "%s: implement RAID set integrity checks", handler);
	return 0;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about the RAID device.
 */
static void template_log(struct lib_context *lc, struct raid_dev *rd)
{
	struct template *template = META(rd, template);

	/* CODEME: implement (use P(), ... macors from format.h */
	log_print(lc, "%s: implement displaying metadata variables", handler);
}
#endif

static struct dmraid_format template_format = {
	.name	= HANDLER,
	.descr	= "(-) Template RAID",
	.format = FMT_RAID,
	.read	= template_read,
	.write	= template_write,
	.group	= template_group,
	.check	= template_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= template_log,
#endif
};

/* Register this format handler with the format core */
int register_template(struct lib_context *lc)
{
	return register_format_handler(lc, &template_format);
}

/* CODEME: Set the RAID device contents up derived from the TEMPLATE ones */
static int setup_rd(struct lib_context *lc, struct raid_dev *rd,
		    struct template *template, struct dev_info *di)
{
	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas->offset = TEMPLATE_CONFIGOFFSET >> 9;
	rd->meta_areas->size = sizeof(*template);
	rd->meta_areas->area = (void*) template;

        rd->di = di;
	rd->fmt = &template_format;

	rd->offset = TEMPLATE_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* CODEME: correct sectors */
	rd->sectors -= rd->sectors & 1;

	rd->status = s_ok; /* CODEME: derive from metadata. */

        if ((rd->name = name(lc, rd, 1)))
		return 1;

	return 0;
}
